"""Definitions of algorithms."""
from .algorithm_base import *
from .inprocess.adv_debiasing import *
from .inprocess.agarwal_reductions import *
from .inprocess.blind import *
from .inprocess.fair_dummies import *
from .inprocess.fairness_wo_demographics import *
from .inprocess.hgr import *
from .inprocess.in_algorithm import *
from .inprocess.in_subprocess import *
from .inprocess.installed_model import *
from .inprocess.kamiran import *
from .inprocess.kamishima import *
from .inprocess.logistic_regression import *
from .inprocess.majority import *
from .inprocess.manual import *
from .inprocess.mlp import *
from .inprocess.oracle import *
from .inprocess.svm import *
from .inprocess.zafar import *
from .postprocess.dp_flip import *
from .postprocess.hardt import *
from .postprocess.post_algorithm import *
from .preprocess.beutel import *
from .preprocess.calders import *
from .preprocess.pre_algorithm import *
from .preprocess.pre_subprocess import *
from .preprocess.upsampler import *
from .preprocess.vfae import *
from .preprocess.zemel import *
